/*
 * Copyright 2025 Squircle CE contributors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.blacksquircle.ui.feature.git.ui.git

import androidx.compose.foundation.layout.Column
import androidx.compose.runtime.Composable
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.tooling.preview.PreviewLightDark
import androidx.navigation.NavController
import com.blacksquircle.ui.core.extensions.daggerViewModel
import com.blacksquircle.ui.core.extensions.showToast
import com.blacksquircle.ui.core.mvi.ViewEvent
import com.blacksquircle.ui.ds.PreviewBackground
import com.blacksquircle.ui.ds.dialog.AlertDialog
import com.blacksquircle.ui.feature.git.R
import com.blacksquircle.ui.feature.git.api.navigation.GitDialog
import com.blacksquircle.ui.feature.git.internal.GitComponent
import com.blacksquircle.ui.feature.git.ui.git.compose.GitAction
import com.blacksquircle.ui.ds.R as UiR

@Composable
internal fun GitScreen(
    navArgs: GitDialog,
    navController: NavController,
    viewModel: GitViewModel = daggerViewModel { context ->
        val component = GitComponent.buildOrGet(context)
        GitViewModel.ParameterizedFactory(navArgs.repository).also(component::inject)
    }
) {
    GitScreen(
        onFetchClicked = viewModel::onFetchClicked,
        onPullClicked = viewModel::onPullClicked,
        onCommitClicked = viewModel::onCommitClicked,
        onPushClicked = viewModel::onPushClicked,
        onCheckoutClicked = viewModel::onCheckoutClicked,
        onBackClicked = viewModel::onBackClicked
    )

    val context = LocalContext.current
    LaunchedEffect(Unit) {
        viewModel.viewEvent.collect { event ->
            when (event) {
                is ViewEvent.Toast -> context.showToast(text = event.message)
                is ViewEvent.Navigation -> navController.navigate(event.screen)
                is ViewEvent.PopBackStack -> navController.popBackStack()
            }
        }
    }
}

@Composable
private fun GitScreen(
    onFetchClicked: () -> Unit = {},
    onPullClicked: () -> Unit = {},
    onCommitClicked: () -> Unit = {},
    onPushClicked: () -> Unit = {},
    onCheckoutClicked: () -> Unit = {},
    onBackClicked: () -> Unit = {}
) {
    AlertDialog(
        title = stringResource(R.string.label_git),
        horizontalPadding = false,
        content = {
            Column {
                GitAction(
                    iconRes = UiR.drawable.ic_autorenew,
                    title = stringResource(R.string.git_fetch_title),
                    subtitle = stringResource(R.string.git_fetch_description),
                    onClick = onFetchClicked,
                )
                GitAction(
                    iconRes = UiR.drawable.ic_tray_arrow_down,
                    title = stringResource(R.string.git_pull_title),
                    subtitle = stringResource(R.string.git_pull_description),
                    onClick = onPullClicked,
                )
                GitAction(
                    iconRes = UiR.drawable.ic_source_commit,
                    title = stringResource(R.string.git_commit_title),
                    subtitle = stringResource(R.string.git_commit_description),
                    onClick = onCommitClicked,
                )
                GitAction(
                    iconRes = UiR.drawable.ic_tray_arrow_up,
                    title = stringResource(R.string.git_push_title),
                    subtitle = stringResource(R.string.git_push_description),
                    onClick = onPushClicked,
                )
                GitAction(
                    iconRes = UiR.drawable.ic_folder_data,
                    title = stringResource(R.string.git_checkout_title),
                    subtitle = stringResource(R.string.git_checkout_description),
                    onClick = onCheckoutClicked,
                )
            }
        },
        dismissButton = stringResource(android.R.string.cancel),
        onDismissClicked = onBackClicked,
        onDismiss = onBackClicked
    )
}

@PreviewLightDark
@Composable
private fun GitScreenPreview() {
    PreviewBackground {
        GitScreen()
    }
}